/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/

package casa.lepton.hub;

import java.io.InputStream;
import java.net.DatagramPacket;

import casa.util.Logger;
import casa.lepton.hub.Beacon;
import casa.lepton.hub.OppNetAdapter;
import casa.lepton.hub.IPND_Beacon;

public class IBRDTN_Adapter
    implements OppNetAdapter {

    // ------------------------------------------------------------
    // Decode the packet as a beacon
    public Beacon getBeacon(DatagramPacket packet)
	throws Exception {
	
	return new IPND_Beacon(packet);
    }

    // ------------------------------------------------------------
    // Decode the sequence of bytes received from an input stream,
    // until the identity of the source of this stream has been
    // identified, then return a StreamHeader that contains the source
    // id, and the bytes that have been read so far
    public StreamHeader getStreamHeader(InputStream in)
	throws Exception {

	// Decoding the contact header according to
	// https://tools.ietf.org/html/draft-irtf-dtnrg-tcp-clayer-02

	byte[] buffer = new byte[64];

	// Expecting a header with a magic cookie in byte[0:4]
	int hlength = in.read(buffer, 0, buffer.length);
	
	// Logger.log("ibrdtn", "<< [" + hlength_ + "] header");

	String cookie = new String(buffer, 0, 4);
	if (! cookie.equals("dtn!")) {
	    Logger.log("ibrdtn", "<< invalid cookie in stream header");
	    throw new Exception("IBRDTN.getStreamHeader(): invalid cookie");
	}
	
	// Expecting the sender's EID length formatted as an SDNV
	// value, starting at byte[8].
	int offset = 8;
	
	int eidlen = (int)SDNV.decode(buffer, offset);
	offset += SDNV.encodingLength(eidlen);

	// Reading the sender's EID (bypassing the prefix 'dtn://')
	String source = new String(buffer, offset+6, eidlen-6);

	Logger.log("ibrdtn", "<< source=" + source);
	byte[] dest = new byte[hlength];
	System.arraycopy(buffer, 0, dest, 0, hlength);
	buffer = null;
	
	return new StreamHeader(source, dest);
    }

    // ------------------------------------------------------------
    // Decode the packet and return the identity of its source
    public String getSource(DatagramPacket packet)
	throws Exception {

	byte[] buffer = packet.getData();
	int offset = packet.getOffset();
	
 	int eidlen = (int)buffer[offset];
	offset++;
	
	// Reading the sender's EID (bypassing the prefix 'dtn://')
	String source = new String(buffer, offset+6, eidlen-6);

	return source;
    }
  
}
