/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/

package casa.lepton.hub;

import java.util.Map;
import java.util.HashMap;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.DatagramPacket;

import casa.util.DataBufferizer;
import casa.lepton.hub.Beacon;
import casa.lepton.hub.SDNV;

public class IPND_Beacon implements Beacon {

    // Implementation according to
    // https://tools.ietf.org/html/draft-irtf-dtnrg-ipnd-01
    
    private byte version_ = 2;
    private byte flags_;
    private short seqnum_;
    private String eid_;
    private String source_;
    
    byte[] bloom_ = null;
    
    private Map<String,String> services_ = new HashMap<String,String>();


    // ------------------------------------------------------------
    public IPND_Beacon(DatagramPacket packet)
	throws Exception {
	decode(packet.getData(), packet.getOffset(), packet.getLength());
    }

    // ------------------------------------------------------------
    public String toString() {

	String result = "IPND_Beacon(version=" + version_
	    + ",flags=0x" + DataBufferizer.hexa(flags_)
	    + ",seqnum=" + seqnum_
	    // + ",eidlen=" + eid_.length()
	    + ",eid=\"" + eid_ + "\""
	    + ",nbsrv=" + services_.size();

	for (String sname: services_.keySet()) {
	    String spar = services_.get(sname);
	    result += "," // ",snlen=" + sname.length() + ",sname=\""
		+ sname + "=\""
		// + ",splen=" + spar.length() + ",spar=\""
		+ spar + "\"";
	}

	result += ",bloom=[" + bloom_.length + "]";

	return result;
    }
      
    // ------------------------------------------------------------
    public String getEID() {
	return eid_;
    }

    // ------------------------------------------------------------
    public BeaconType getBeaconType() {

	// IPND beacons are all "HELLO" beacons: there is no way for a
	// host to announce it is leaving the network by broadcasting
	// a "BYE" beacon
	return BeaconType.HELLO;
    }

    // ------------------------------------------------------------
    public boolean hasGossipingType(GossipingType type) {

	return services_.containsKey(gossipingTypeToName(type));
    }

    // ------------------------------------------------------------
    public String getSource() {

	return source_;
    }

    // ------------------------------------------------------------
    public String getDestination() {

	// IPND beacons are always broadcast, so no specific
	// destination is ever specified
	return null;
    }

    // ------------------------------------------------------------
    public String getService(String sname) {
	return services_.get(sname);
    }

    // ------------------------------------------------------------
    public void setService(String sname, String spart) {
	services_.put(sname, spart);
    }

    // ------------------------------------------------------------
    private String gossipingTypeToName(GossipingType type) {

	switch(type) {
	case TCP:
	    return "tcpcl";
	case UDP:
	    return "udpcl";
	default:
	    return null;
	}
    }

    // ------------------------------------------------------------
    private GossipingType gossipingNameToType(String type) {

	if (type.equals("tcpcl"))
	    return GossipingType.TCP;
	if (type.equals("udpcl"))
	    return GossipingType.UDP;
	return null;
    }
    
    // ------------------------------------------------------------
    public InetSocketAddress getGossipingAddress(GossipingType type) {

	InetSocketAddress result = null;
	String spar = services_.get(gossipingTypeToName(type));

	if (spar == null)
	    return null;

	try {
	    int idx1 = spar.indexOf("port=");
	    int idx2 = spar.indexOf(";", idx1);
	    String tmp = spar.substring(idx1 + 5, idx2);

	    int port = Integer.parseInt(tmp);

	    idx1 = spar.indexOf("ip=");
	    if (idx1 != -1) {
		idx2 = spar.indexOf(";", idx1);
		tmp = spar.substring(idx1 + 3, idx2);
		result = new InetSocketAddress(tmp, port);
	    }
	    else
		result = new InetSocketAddress((InetAddress)null, port);
	}
	catch (Exception e) {}

	return result;
    }

    // ------------------------------------------------------------
    public void setGossipingAddress(GossipingType type,
				    InetSocketAddress addr) {

	String serv = gossipingTypeToName(type);
	if (serv == null)
	    return;

	if (addr.getAddress().isAnyLocalAddress())
	    services_.put(serv, "port=" + addr.getPort() + ";");
	else
	    services_.put(serv, "ip=" + addr.getAddress().getHostAddress()
			  + ";port=" + addr.getPort() + ";");
    }

    // ------------------------------------------------------------
    public void decode(byte[] buffer, int offset, int length)
	throws Exception {

	// System.out.println("beacon: << " +
	// 	   DataBufferizer.bufferToString(buffer, offset, length));

	version_ = DataBufferizer.readByte(buffer, offset);
	offset++;
	flags_   = DataBufferizer.readByte(buffer, offset);
	offset++;
	seqnum_ = DataBufferizer.readShort(buffer, offset);
	offset+= 2;
	long eidlen = SDNV.decode(buffer, offset);
	offset += SDNV.encodingLength(eidlen);
	eid_ = new String(buffer, offset, (int)eidlen);
	offset += eidlen;

	int idx1 = eid_.indexOf("dtn://");
	source_ = eid_.substring(idx1+6);
	
	long nbsrv = SDNV.decode(buffer, offset);
	offset += SDNV.encodingLength(nbsrv);
	
	services_.clear();
	for (long i=0; i<nbsrv; i++) {
	    long snlen = SDNV.decode(buffer, offset);
	    offset += SDNV.encodingLength(snlen);
	    String sname = new String(buffer, offset, (int)snlen);
	    offset += snlen;
	    long splen = SDNV.decode(buffer, offset);
	    offset += SDNV.encodingLength(splen);
	    String spar = new String(buffer, offset, (int)splen);
	    offset += splen;
	    services_.put(sname, spar);

	}
	
	int bloom_length = length - offset;
	bloom_ = new byte[bloom_length];
	java.lang.System.arraycopy(buffer, offset, bloom_, 0, bloom_length);
    }

    // ------------------------------------------------------------
    public int encode(byte[] buffer, int offset)
	throws Exception {

	DataBufferizer.writeByte(version_, buffer, offset);
	offset++;
	DataBufferizer.writeByte(flags_, buffer, offset);
	offset++;
	DataBufferizer.writeShort(seqnum_, buffer, offset);
	offset+= 2;
	byte[] tmp = eid_.getBytes();
	int eidlen = tmp.length;
	SDNV.encode(eidlen, buffer, offset);
	offset += SDNV.encodingLength(eidlen);
	java.lang.System.arraycopy(tmp, 0, buffer, offset, eidlen);
	offset += eidlen;
	int nbsrv = services_.size();
	SDNV.encode(nbsrv, buffer, offset);
	offset += SDNV.encodingLength(nbsrv);
	
	for (String sname: services_.keySet()) {
	    String spar = services_.get(sname);
	    
	    tmp = sname.getBytes();
	    int snlen = tmp.length;
	    SDNV.encode(snlen, buffer, offset);
	    offset += SDNV.encodingLength(snlen);
	    java.lang.System.arraycopy(tmp, 0, buffer, offset, snlen);
	    offset += snlen;
	    
	    tmp = spar.getBytes();
	    int splen = tmp.length;
	    SDNV.encode(splen, buffer, offset);
	    offset += SDNV.encodingLength(splen);
	    java.lang.System.arraycopy(tmp, 0, buffer, offset, splen);
	    offset += splen;
	}

	java.lang.System.arraycopy(bloom_, 0, buffer, offset, bloom_.length);
	offset += bloom_.length;

	// System.out.println("beacon: >> " +
	// 	   DataBufferizer.bufferToString(buffer, 0, offset));

	return offset;
    }
	    
    // ------------------------------------------------------------
    public static void main(String[] args) throws Exception {

	// Used only to test SDNV encoding and decoding
	
	for (int i=0; i<args.length; i++) {
	    Long v = Long.parseLong(args[i]);
	    System.out.println("SDNV.encodingLength(" + v + ")=" + SDNV.encodingLength(v));
	}

	byte[] buffer = new byte[512];
	for (int i=0; i<args.length; i++) {
	    Long v = Long.parseLong(args[i]);
	    int length = SDNV.encode(v, buffer, 0);
	    System.out.println("SDNV(" + v + ")="
			       + DataBufferizer.bufferToString(buffer, 0, length));
	}
    }
    
}
